package sun.font;

import java.io.File;
import sun.java2d.DisposerRecord;
import java.io.UnsupportedEncodingException;
import java.nio.BufferUnderflowException;
import java.nio.ByteOrder;
import java.io.IOException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.FileChannel;
import java.security.AccessController;
import java.io.FileNotFoundException;
import java.security.PrivilegedAction;
import java.io.RandomAccessFile;
import java.nio.MappedByteBuffer;
import java.nio.ByteBuffer;
import sun.java2d.Disposer;
import java.awt.FontFormatException;
import java.util.HashSet;
import java.util.HashMap;
import java.lang.ref.WeakReference;

public class Type1Font extends FileFont
{
    WeakReference bufferRef;
    private String psName;
    private static HashMap styleAbbreviationsMapping;
    private static HashSet styleNameTokes;
    private static final int PSEOFTOKEN = 0;
    private static final int PSNAMETOKEN = 1;
    private static final int PSSTRINGTOKEN = 2;
    
    public Type1Font(String s, Object o) throws FontFormatException {
        this(s, o, false);
    }
    
    public Type1Font(String s, Object o, boolean b) throws FontFormatException {
        super(s, o);
        this.bufferRef = new WeakReference(null);
        this.psName = null;
        this.fontRank = 4;
        this.checkedNatives = true;
        try {
            this.verify();
        }
        catch (Throwable t) {
            if (b) {
                Disposer.addObjectRecord(this.bufferRef, new T1DisposerRecord(s));
                this.bufferRef = null;
            }
            if ((t instanceof FontFormatException)) {
                throw (FontFormatException)t;
            }
            throw new FontFormatException("Unexpected runtime exception.");
        }
    }
    
    private synchronized ByteBuffer getBuffer() throws FontFormatException {
        MappedByteBuffer map = (MappedByteBuffer)this.bufferRef.get();
        if (map == null) {
            try {
                FileChannel channel = ((RandomAccessFile)AccessController.doPrivileged(new PrivilegedAction() {
                    @Override
                    public Object run() {
                        try {
                            return new RandomAccessFile(Type1Font.this.platName, "r");
                        }
                        catch (FileNotFoundException ex) {
                            return null;
                        }
                    }
                })).getChannel();
                this.fileSize = (int)channel.size();
                map = channel.map(FileChannel.MapMode.READ_ONLY, 0L, (long)this.fileSize);
                map.position(0);
                this.bufferRef = new WeakReference(map);
                channel.close();
            }
            catch (NullPointerException ex) {
                throw new FontFormatException(ex.toString());
            }
            catch (ClosedChannelException ex2) {
                Thread.interrupted();
                return this.getBuffer();
            }
            catch (IOException ex3) {
                throw new FontFormatException(ex3.toString());
            }
        }
        return map;
    }
    
    @Override
    protected void close() {
    }
    
    void readFile(ByteBuffer byteBuffer) {
        RandomAccessFile randomAccessFile = null;
        try {
            randomAccessFile = (RandomAccessFile)AccessController.doPrivileged(new PrivilegedAction() {
                @Override
                public Object run() {
                    try {
                        return new RandomAccessFile(Type1Font.this.platName, "r");
                    }
                    catch (FileNotFoundException ex) {
                        return null;
                    }
                }
            });
            FileChannel channel = randomAccessFile.getChannel();
            while (byteBuffer.remaining() > 0 && channel.read(byteBuffer) != -1) {}
        }
        catch (NullPointerException ex) {}
        catch (ClosedChannelException ex2) {
            try {
                if (randomAccessFile != null) {
                    randomAccessFile.close();
                    randomAccessFile = null;
                }
            }
            catch (IOException ex3) {}
            Thread.interrupted();
            this.readFile(byteBuffer);
        }
        catch (IOException ex4) {}
        finally {
            if (randomAccessFile != null) {
                try {
                    randomAccessFile.close();
                }
                catch (IOException ex5) {}
            }
        }
    }
    
    public synchronized ByteBuffer readBlock(int fileSize, int n) {
        try {
            ByteBuffer buffer = this.getBuffer();
            if (fileSize > this.fileSize) {
                fileSize = this.fileSize;
            }
            buffer.position(fileSize);
            return buffer.slice();
        }
        catch (FontFormatException ex) {
            return null;
        }
    }
    
    private void verify() throws FontFormatException {
        ByteBuffer buffer = this.getBuffer();
        if (buffer.capacity() < 6) {
            throw new FontFormatException("short file");
        }
        int n = buffer.get(0) & 0xFF;
        if ((buffer.get(0) & 0xFF) == 0x80) {
            this.verifyPFB(buffer);
            buffer.position(6);
        }
        else {
            this.verifyPFA(buffer);
            buffer.position(0);
        }
        this.initNames(buffer);
        if (this.familyName == null || this.fullName == null) {
            throw new FontFormatException("Font name not found");
        }
        this.setStyle();
        return;
    }
    
    public int getFileSize() {
        if (this.fileSize == 0) {
            try {
                this.getBuffer();
            }
            catch (FontFormatException ex) {}
        }
        return this.fileSize;
    }
    
    private void verifyPFA(ByteBuffer byteBuffer) throws FontFormatException {
        if (byteBuffer.getShort() != 9505) {
            throw new FontFormatException("bad pfa font");
        }
        return;
    }
    
    private void verifyPFB(ByteBuffer byteBuffer) throws FontFormatException {
        int n = 0;
        while (true) {
            try {
                int n2 = byteBuffer.getShort(n) & 0xFFFF;
                if (n2 == 32769 || n2 == 32770) {
                    byteBuffer.order(ByteOrder.LITTLE_ENDIAN);
                    int int1 = byteBuffer.getInt(n + 2);
                    byteBuffer.order(ByteOrder.BIG_ENDIAN);
                    if (int1 <= 0) {
                        throw new FontFormatException("bad segment length");
                    }
                    n += int1 + 6;
                    continue;
                }
                else {
                    if (n2 == 32771) {
                        return;
                    }
                    throw new FontFormatException("bad pfb file");
                }
            }
            catch (BufferUnderflowException ex) {
                throw new FontFormatException(ex.toString());
            }
            catch (Exception ex2) {
                throw new FontFormatException(ex2.toString());
            }
        }
    }
    
    private void initNames(ByteBuffer byteBuffer) throws FontFormatException {
        int n = 0;
        String anObject = null;
        try {
            while ((this.fullName == null || this.familyName == null || this.psName == null || anObject == null) && n == 0) {
                int nextTokenType = this.nextTokenType(byteBuffer);
                if (nextTokenType == 1) {
                    if (byteBuffer.get(byteBuffer.position()) == 70) {
                        String simpleToken = this.getSimpleToken(byteBuffer);
                        if ("FullName".equals(simpleToken)) {
                            if (this.nextTokenType(byteBuffer) == 2) {
                                this.fullName = this.getString(byteBuffer);
                            }
                            else {
                                continue;
                            }
                        }
                        else if ("FamilyName".equals(simpleToken)) {
                            if (this.nextTokenType(byteBuffer) == 2) {
                                this.familyName = this.getString(byteBuffer);
                            }
                            else {
                                continue;
                            }
                        }
                        else if ("FontName".equals(simpleToken)) {
                            if (this.nextTokenType(byteBuffer) == 1) {
                                this.psName = this.getSimpleToken(byteBuffer);
                            }
                            else {
                                continue;
                            }
                        }
                        else if ("FontType".equals(simpleToken)) {
                            String simpleToken2 = this.getSimpleToken(byteBuffer);
                            if ("def".equals(this.getSimpleToken(byteBuffer))) {
                                anObject = simpleToken2;
                            }
                            else {
                                continue;
                            }
                        }
                        else {
                            continue;
                        }
                    }
                    else {
                        while (byteBuffer.get() > 32) {}
                    }
                }
                else if (nextTokenType == 0) {
                    n = 1;
                }
                else {
                    continue;
                }
            }
        }
        catch (Exception ex) {
            throw new FontFormatException(ex.toString());
        }
        if (!"1".equals(anObject)) {
            throw new FontFormatException("Unsupported font type");
        }
        if (this.psName == null) {
            byteBuffer.position(0);
            if (byteBuffer.getShort() != 9505) {
                byteBuffer.position(8);
            }
            String simpleToken3 = this.getSimpleToken(byteBuffer);
            if (!simpleToken3.startsWith("FontType1-") && !simpleToken3.startsWith("PS-AdobeFont-")) {
                throw new FontFormatException("Unsupported font format [" + simpleToken3 + "]");
            }
            this.psName = this.getSimpleToken(byteBuffer);
        }
        if (n != 0) {
            if (this.fullName != null) {
                this.familyName = this.fullName2FamilyName(this.fullName);
            }
            else if (this.familyName != null) {
                this.fullName = this.familyName;
            }
            else {
                this.fullName = this.psName2FullName(this.psName);
                this.familyName = this.psName2FamilyName(this.psName);
            }
        }
        return;
    }
    
    private String fullName2FamilyName(String s) {
        int i = s.length();
        while (i > 0) {
            int index;
            for (index = i - 1; index > 0 && s.charAt(index) != ' '; --index) {}
            if (!this.isStyleToken(s.substring(index + 1, i))) {
                return s.substring(0, i);
            }
            i = index;
        }
        return s;
    }
    
    private String expandAbbreviation(String s) {
        if (Type1Font.styleAbbreviationsMapping.containsKey(s)) {
            return (String)Type1Font.styleAbbreviationsMapping.get(s);
        }
        return s;
    }
    
    private boolean isStyleToken(String o) {
        return Type1Font.styleNameTokes.contains(o);
    }
    
    private String psName2FullName(String s) {
        int index = s.indexOf("-");
        if (index >= 0) {
            return this.expandName(s.substring(0, index), false) + " " + this.expandName(s.substring(index + 1), true);
        }
        return this.expandName(s, false);
    }
    
    private String psName2FamilyName(String substring) {
        if (substring.indexOf("-") > 0) {
            substring = substring.substring(0, substring.indexOf("-"));
        }
        return this.expandName(substring, false);
    }
    
    private int nextCapitalLetter(String s, int n) {
        while (n >= 0 && n < s.length()) {
            if (s.charAt(n) >= 'A' && s.charAt(n) <= 'Z') {
                return n;
            }
            ++n;
        }
        return -1;
    }
    
    private String expandName(String s, boolean b) {
        StringBuffer sb = new StringBuffer(s.length() + 10);
        int n;
        for (int i = 0; i < s.length(); i = n) {
            n = this.nextCapitalLetter(s, i + 1);
            if (n < 0) {
                n = s.length();
            }
            if (i != 0) {
                sb.append(" ");
            }
            if (b) {
                sb.append(this.expandAbbreviation(s.substring(i, n)));
            }
            else {
                sb.append(s.substring(i, n));
            }
        }
        return sb.toString();
    }
    
    private byte skip(ByteBuffer byteBuffer) {
        int i = byteBuffer.get();
        while (i == 37) {
            do {
                i = byteBuffer.get();
                if (i != 13) {
                    continue;
                }
                break;
            } while (i != 10);
        }
        while (i <= 32) {
            i = byteBuffer.get();
        }
        return i;
    }
    
    private int nextTokenType(ByteBuffer byteBuffer) {
        try {
            int i = this.skip(byteBuffer);
            while (i != 47) {
                if (i == 40) {
                    return 2;
                }
                if (i == 13 || i == 10) {
                    i = this.skip(byteBuffer);
                }
                else {
                    i = byteBuffer.get();
                }
            }
            return 1;
        }
        catch (BufferUnderflowException ex) {
            return 0;
        }
    }
    
    private String getSimpleToken(ByteBuffer byteBuffer) {
        while (byteBuffer.get() <= 32) {}
        int n = byteBuffer.position() - 1;
        while (byteBuffer.get() > 32) {}
        byte[] bytes = new byte[byteBuffer.position() - n - 1];
        byteBuffer.position(n);
        byteBuffer.get(bytes);
        try {
            return new String(bytes, "US-ASCII");
        }
        catch (UnsupportedEncodingException ex) {
            return new String(bytes);
        }
    }
    
    private String getString(ByteBuffer byteBuffer) {
        int position = byteBuffer.position();
        while (byteBuffer.get() != 41) {}
        byte[] bytes = new byte[byteBuffer.position() - position - 1];
        byteBuffer.position(position);
        byteBuffer.get(bytes);
        try {
            return new String(bytes, "US-ASCII");
        }
        catch (UnsupportedEncodingException ex) {
            return new String(bytes);
        }
    }
    
    @Override
    public String getPostscriptName() {
        return this.psName;
    }
    
    @Override
    protected synchronized FontScaler getScaler() {
        if (this.scaler == null) {
            this.scaler = FontScaler.getScaler(this, 0, false, this.fileSize);
        }
        return this.scaler;
    }
    
    @Override
    CharToGlyphMapper getMapper() {
        if (this.mapper == null) {
            this.mapper = new Type1GlyphMapper(this);
        }
        return this.mapper;
    }
    
    @Override
    public int getNumGlyphs() {
        try {
            return this.getScaler().getNumGlyphs();
        }
        catch (FontScalerException ex) {
            this.scaler = FontScaler.getNullScaler();
            return this.getNumGlyphs();
        }
    }
    
    @Override
    public int getMissingGlyphCode() {
        try {
            return this.getScaler().getMissingGlyphCode();
        }
        catch (FontScalerException ex) {
            this.scaler = FontScaler.getNullScaler();
            return this.getMissingGlyphCode();
        }
    }
    
    public int getGlyphCode(char c) {
        try {
            return this.getScaler().getGlyphCode(c);
        }
        catch (FontScalerException ex) {
            this.scaler = FontScaler.getNullScaler();
            return this.getGlyphCode(c);
        }
    }
    
    @Override
    public String toString() {
        return "** Type1 Font: Family=" + this.familyName + " Name=" + this.fullName + " style=" + this.style + " fileName=" + this.getPublicFileName();
    }
    
    static {
        Type1Font.styleAbbreviationsMapping = new HashMap();
        Type1Font.styleNameTokes = new HashSet();
        String[] array = { "Black", "Bold", "Book", "Demi", "Heavy", "Light", "Meduium", "Nord", "Poster", "Regular", "Super", "Thin", "Compressed", "Condensed", "Compact", "Extended", "Narrow", "Inclined", "Italic", "Kursiv", "Oblique", "Upright", "Sloped", "Semi", "Ultra", "Extra", "Alternate", "Alternate", "Deutsche Fraktur", "Expert", "Inline", "Ornaments", "Outline", "Roman", "Rounded", "Script", "Shaded", "Swash", "Titling", "Typewriter" };
        String[] array2 = { "Blk", "Bd", "Bk", "Dm", "Hv", "Lt", "Md", "Nd", "Po", "Rg", "Su", "Th", "Cm", "Cn", "Ct", "Ex", "Nr", "Ic", "It", "Ks", "Obl", "Up", "Sl", "Sm", "Ult", "X", "A", "Alt", "Dfr", "Exp", "In", "Or", "Ou", "Rm", "Rd", "Scr", "Sh", "Sw", "Ti", "Typ" };
        String[] array3 = { "Black", "Bold", "Book", "Demi", "Heavy", "Light", "Medium", "Nord", "Poster", "Regular", "Super", "Thin", "Compressed", "Condensed", "Compact", "Extended", "Narrow", "Inclined", "Italic", "Kursiv", "Oblique", "Upright", "Sloped", "Slanted", "Semi", "Ultra", "Extra" };
        for (int i = 0; i < array.length; ++i) {
            Type1Font.styleAbbreviationsMapping.put(array2[i], array[i]);
        }
        for (int j = 0; j < array3.length; ++j) {
            Type1Font.styleNameTokes.add(array3[j]);
        }
    }
    
    private static class T1DisposerRecord implements DisposerRecord
    {
        String fileName;
        
        T1DisposerRecord(String fileName) {
            this.fileName = null;
            this.fileName = fileName;
        }
        
        @Override
        public synchronized void dispose() {
            AccessController.doPrivileged(new PrivilegedAction() {
                @Override
                public Object run() {
                    if (Type1Font.T1DisposerRecord.this.fileName != null) {
                        new File(Type1Font.T1DisposerRecord.this.fileName).delete();
                    }
                    return null;
                }
            });
        }
    }
}
